#!/usr/bin/python
#dhnview.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#

if __name__ == "__main__":
    import p2p.dhnview
    p2p.dhnview.main()


